@@ -16,6 +16,10 @@ from aiodns import error as aiodns_error
 from aiopg import pool
 from psycopg2 import errors, extras
 from tornado import ioloop, web
+try:
+    import problemdetails
+except ImportError:  # pragma: nocover
+    problemdetails = None
 
 LOGGER = logging.getLogger('sprockets-postgres')
 
@@ -110,7 +114,7 @@ class PostgresConnector:
     """
     def __init__(self,
                  cursor: aiopg.Cursor,
-                 on_error: typing.Callable,
+                 on_error: typing.Optional[typing.Callable] = None,
                  on_duration: typing.Optional[typing.Callable] = None,
                  timeout: Timeout = None):
         self.cursor = cursor
@@ -246,9 +250,10 @@ class PostgresConnector:
         try:
             await method(**kwargs)
         except (asyncio.TimeoutError, psycopg2.Error) as err:
-            exc = self._on_error(metric_name, err)
-            if exc:
-                raise exc
+            if self._on_error:
+                err = self._on_error(metric_name, err)
+            if err:
+                raise err
         else:
             results = await self._query_results()
             if self._on_duration:
@@ -258,16 +263,31 @@ class PostgresConnector:
 
     async def _query_results(self) -> QueryResult:
         count, row, rows = self.cursor.rowcount, None, None
+
+        def _on_programming_error(err: psycopg2.ProgrammingError) -> None:
+            # Should always be empty in this context
+            if err.pgcode is not None:  # pragma: nocover
+                LOGGER.warning(
+                    'Unexpected value for ProgrammingError(%s).pgcode: %r',
+                    err, err.pgcode)
+
+        # This logic exists so that we can wrap the results and quickly free
+        # the cursor / connection. We do not know if the query was a SELECT,
+        # INSERT, UPDATE, or DELETE and we may impact multiple rows, but not
+        # have any data to retrieve. The ProgrammingErrror is raised in this
+        # context. It is raised in other contexts, such as when there is an
+        # error in the query, but that will be caught on L248
+
         if self.cursor.rowcount == 1:
             try:
                 row = dict(await self.cursor.fetchone())
-            except psycopg2.ProgrammingError:
-                pass
+            except psycopg2.ProgrammingError as exc:
+                _on_programming_error(exc)
         elif self.cursor.rowcount > 1:
             try:
                 rows = [dict(row) for row in await self.cursor.fetchall()]
-            except psycopg2.ProgrammingError:
-                pass
+            except psycopg2.ProgrammingError as exc:
+                _on_programming_error(exc)
         return QueryResult(count, row, rows)
 
 
@@ -298,7 +318,8 @@ class ApplicationMixin:
 
     @contextlib.asynccontextmanager
     async def postgres_connector(self,
-                                 on_error: typing.Callable,
+                                 on_error: typing.Optional[
+                                     typing.Callable] = None,
                                  on_duration: typing.Optional[
                                      typing.Callable] = None,
                                  timeout: Timeout = None,
@@ -340,26 +361,25 @@ class ApplicationMixin:
                         timeout=timeout) as cursor:
                     yield PostgresConnector(
                         cursor, on_error, on_duration, timeout)
-        except (asyncio.TimeoutError, psycopg2.Error) as err:
-            message = str(err)
+        except (asyncio.TimeoutError, psycopg2.OperationalError) as err:
             if isinstance(err, psycopg2.OperationalError) and _attempt == 1:
                 LOGGER.critical('Disconnected from Postgres: %s', err)
-                retry = True
                 if not self._postgres_reconnect.locked():
                     async with self._postgres_reconnect:
-                        retry = await self._postgres_connect()
-                if retry:
-                    await self._postgres_connected.wait()
-                    async with self.postgres_connector(
-                            on_error, on_duration, timeout,
-                            _attempt + 1) as connector:
-                        yield connector
-                    return
-                message = 'disconnected'
-            exc = on_error('postgres_connector', ConnectionException(message))
+                        if await self._postgres_connect():
+                            await self._postgres_connected.wait()
+                            async with self.postgres_connector(
+                                    on_error, on_duration, timeout,
+                                    _attempt + 1) as connector:
+                                yield connector
+                            return
+            if on_error is None:
+                raise ConnectionException(str(err))
+            exc = on_error(
+                'postgres_connector', ConnectionException(str(err)))
             if exc:
                 raise exc
-            else:   # postgres_status.on_error does not return an exception
+            else:  # postgres_status.on_error does not return an exception
                 yield None
 
     async def postgres_status(self) -> dict:
@@ -433,7 +453,9 @@ class ApplicationMixin:
         if self._postgres_pool:
             self._postgres_pool.close()
 
-        LOGGER.debug('Connecting to %s', url)
+        safe_url = self._obscure_url_password(url)
+        LOGGER.debug('Connecting to %s', safe_url)
+
         try:
             self._postgres_pool = await pool.Pool.from_pool_fill(
                 url,
@@ -464,13 +486,25 @@ class ApplicationMixin:
                         DEFAULT_POSTGRES_CONNECTION_TTL)))
         except (psycopg2.OperationalError,
                 psycopg2.Error) as error:  # pragma: nocover
-            LOGGER.warning('Error connecting to PostgreSQL on startup: %s',
-                           error)
+            LOGGER.warning(
+                'Error connecting to PostgreSQL on startup with %s: %s',
+                safe_url, error)
             return False
         self._postgres_connected.set()
         LOGGER.debug('Connected to Postgres')
         return True
 
+    @staticmethod
+    def _obscure_url_password(url):
+        """Generate log safe url with password obscured."""
+        parsed = parse.urlparse(url)
+        if parsed.password:
+            netloc = '{}:*****@{}:{}'.format(parsed.username,
+                                             parsed.hostname,
+                                             parsed.port)
+            url = parse.urlunparse(parsed._replace(netloc=netloc))
+        return url
+
     async def _postgres_on_start(self,
                                  _app: web.Application,
                                  loop: ioloop.IOLoop):
@@ -577,8 +611,8 @@ class RequestHandlerMixin:
 
         """
         async with self.application.postgres_connector(
-                self._on_postgres_error,
-                self._on_postgres_timing,
+                self.on_postgres_error,
+                self.on_postgres_timing,
                 timeout) as connector:
             return await connector.callproc(
                 name, parameters, metric_name, timeout=timeout)
@@ -616,8 +650,8 @@ class RequestHandlerMixin:
 
         """
         async with self.application.postgres_connector(
-                self._on_postgres_error,
-                self._on_postgres_timing,
+                self.on_postgres_error,
+                self.on_postgres_timing,
                 timeout) as connector:
             return await connector.execute(
                 sql, parameters, metric_name, timeout=timeout)
@@ -663,16 +697,22 @@ class RequestHandlerMixin:
 
         """
         async with self.application.postgres_connector(
-                self._on_postgres_error,
-                self._on_postgres_timing,
+                self.on_postgres_error,
+                self.on_postgres_timing,
                 timeout) as connector:
             async with connector.transaction():
                 yield connector
 
-    def _on_postgres_error(self,
-                           metric_name: str,
-                           exc: Exception) -> typing.Optional[Exception]:
-        """Override for different error handling behaviors
+    def on_postgres_error(self,
+                          metric_name: str,
+                          exc: Exception) -> typing.Optional[Exception]:
+        """Invoked when an error occurs when executing a query
+
+        If `tornado-problem-details` is available,
+        :exc:`problemdetails.Problem` will be raised instead of
+        :exc:`tornado.web.HTTPError`.
+
+        Override for different error handling behaviors.
 
         Return an exception if you would like for it to be raised, or swallow
         it here.
@@ -682,18 +722,30 @@ class RequestHandlerMixin:
                      exc.__class__.__name__, self.__class__.__name__,
                      metric_name, str(exc).split('\n')[0])
         if isinstance(exc, ConnectionException):
+            if problemdetails:
+                raise problemdetails.Problem(
+                    status_code=503, title='Database Connection Error')
             raise web.HTTPError(503, reason='Database Connection Error')
         elif isinstance(exc, asyncio.TimeoutError):
+            if problemdetails:
+                raise problemdetails.Problem(
+                    status_code=500, title='Query Timeout')
             raise web.HTTPError(500, reason='Query Timeout')
         elif isinstance(exc, errors.UniqueViolation):
+            if problemdetails:
+                raise problemdetails.Problem(
+                    status_code=409, title='Unique Violation')
             raise web.HTTPError(409, reason='Unique Violation')
         elif isinstance(exc, psycopg2.Error):
+            if problemdetails:
+                raise problemdetails.Problem(
+                    status_code=500, title='Database Error')
             raise web.HTTPError(500, reason='Database Error')
         return exc
 
-    def _on_postgres_timing(self,
-                            metric_name: str,
-                            duration: float) -> None:
+    def on_postgres_timing(self,
+                           metric_name: str,
+                           duration: float) -> None:
         """Override for custom metric recording. As a default behavior it will
         attempt to detect `sprockets-influxdb
         <https://sprockets-influxdb.readthedocs.io/>`_ and
@@ -713,3 +765,21 @@ class RequestHandlerMixin:
         else:
             LOGGER.debug('Postgres query %s duration: %s',
                          metric_name, duration)
+
+
+class StatusRequestHandler(web.RequestHandler):
+    """A RequestHandler that can be used to expose API health or status"""
+
+    async def get(self, *_args, **_kwarg):
+        postgres = await self.application.postgres_status()
+        if not postgres['available']:
+            self.set_status(503)
+        self.write({
+            'application': self.settings.get('service', 'unknown'),
+            'environment': self.settings.get('environment', 'unknown'),
+            'postgres': {
+                'pool_free': postgres['pool_free'],
+                'pool_size': postgres['pool_size']
+            },
+            'status': 'ok' if postgres['available'] else 'unavailable',
+            'version': self.settings.get('version', 'unknown')})
